from six.moves import urllib_parse as urlparse

import jsonpointer

from flex.validation.common import (
    validate_object,
)


class LazyReferenceValidator(object):
    """
    This class acts as a lazy validator for references in schemas to prevent an
    infinite recursion error when a schema references itself, or there is a
    reference loop between more than one schema.

    The validator is only constructed if validator is needed.
    """
    validators_constructor = None

    def __init__(self, reference, context):
        if self.validators_constructor is None:
            raise NotImplementedError(
                "Subclasses of LazyReferenceValidator must specify a "
                "`validators_constructor` function"
            )

        self.reference_fragment = urlparse.urlparse(reference).fragment
        # TODO: something better than this which potentiall raises a
        # JsonPointerException
        jsonpointer.resolve_pointer(context, self.reference_fragment)
        self.reference = reference
        self.context = context

    def __call__(self, value, **kwargs):
        return validate_object(
            value,
            schema=self.schema,
            **kwargs
        )

    @property
    def schema(self):
        return jsonpointer.resolve_pointer(self.context, self.reference_fragment)

    @property
    def validators(self):
        return self.validators_constructor(
            self.schema,
            self.context,
        )

    def items(self):
        return self.validators.items()



import functools

from flex.datastructures import (
    ValidationDict,
    ValidationList,
)
from flex.constants import (
    OBJECT,
)
from flex.validation.common import (
    generate_object_validator,
    apply_validator_to_object,
)
from flex.validation.utils import (
    generate_any_validator,
)
from flex.loading.common.reference import (
    reference_object_validator,
)
from .single import (
    single_response_validator,
)


responses_schema = {
    'type': OBJECT,
}


field_validators = ValidationDict()

field_validators.add_property_validator(
    'default',
    generate_any_validator(
        referenceObject=reference_object_validator,
        responseObject=single_response_validator,
    ),
)


non_field_validators = ValidationList()
non_field_validators.add_validator(
    functools.partial(
        apply_validator_to_object,
        validator=generate_any_validator(
            referenceObject=reference_object_validator,
            responseObject=single_response_validator,
        ),
    )
)


responses_validator = generate_object_validator(
    schema=responses_schema,
    field_validators=field_validators,
    non_field_validators=non_field_validators,
)



from six.moves import urllib_parse as urlparse

import jsonpointer

from flex.exceptions import (
    ValidationError,
)
from flex.datastructures import (
    ValidationDict,
)
from flex.error_messages import MESSAGES
from flex.constants import (
    OBJECT,
    STRING,
)
from flex.decorators import (
    skip_if_not_of_type,
)
from flex.validation.common import (
    generate_object_validator,
)

reference_object_schema = {
    'type': OBJECT,
    'additionalProperties': False,
    'required': [
        '$ref',
    ],
    'properties': {
        '$ref': {
            'type': STRING,
        },
    },
}


@skip_if_not_of_type(STRING)
def validate_reference_pointer(reference, context, **kwargs):
    parts = urlparse.urlparse(reference)
    if any((parts.scheme, parts.netloc, parts.path, parts.params, parts.query)):
        raise ValidationError(
            MESSAGES['reference']['unsupported'].format(reference),
        )

    try:
        jsonpointer.resolve_pointer(context, parts.fragment)
    except jsonpointer.JsonPointerException:
        raise ValidationError(
            MESSAGES['reference']['undefined'].format(reference),
        )


non_field_validators = ValidationDict()
non_field_validators.add_property_validator('$ref', validate_reference_pointer)

reference_object_validator = generate_object_validator(
    schema=reference_object_schema,
    non_field_validators=non_field_validators,
)



import functools
from flex.datastructures import (
    ValidationList,
)
from flex.constants import (
    ARRAY,
)
from flex.validation.common import (
    generate_object_validator,
    apply_validator_to_array,
)
from flex.validation.utils import (
    generate_any_validator,
)
from flex.loading.common.reference import (
    reference_object_validator,
)
from .single import (
    single_parameter_validator,
)


parameters_schema = {
    'type': ARRAY,
}

parameters_non_field_validators = ValidationList()
parameters_non_field_validators.add_validator(
    functools.partial(
        apply_validator_to_array,
        validator=generate_any_validator(
            referenceObject=reference_object_validator,
            parameterObject=single_parameter_validator,
        ),
    ),
)

parameters_validator = generate_object_validator(
    schema=parameters_schema,
    non_field_validators=parameters_non_field_validators,
)



import functools
from flex.datastructures import (
    ValidationList,
)
from flex.constants import (
    OBJECT,
)
from flex.validation.common import (
    generate_object_validator,
    apply_validator_to_object,
)
from .single import (
    single_parameter_validator,
)


parameters_schema = {
    'type': OBJECT,
}

parameters_non_field_validators = ValidationList()
parameters_non_field_validators.add_validator(
    functools.partial(apply_validator_to_object, validator=single_parameter_validator),
)

parameters_validator = generate_object_validator(
    schema=parameters_schema,
    non_field_validators=parameters_non_field_validators,
)



import functools

from flex.exceptions import (
    ValidationError,
    ErrorDict,
)


def any_validator(obj, validators, **kwargs):
    """
    Attempt multiple validators on an object.

    - If any pass, then all validation passes.
    - Otherwise, raise all of the errors.
    """
    if not len(validators) > 1:
        raise ValueError(
            "any_validator requires at least 2 validator.  Only got "
            "{0}".format(len(validators))
        )
    errors = ErrorDict()
    for key, validator in validators.items():
        try:
            validator(obj, **kwargs)
        except ValidationError as err:
            errors[key] = err.detail
        else:
            break
    else:
        if len(errors) == 1:
            # Special case for a single error.  Just raise it as if it was the
            # only validator run.
            error = errors.values()[0]
            raise ValidationError(error)
        else:
            # Raise all of the errors with the key namespaces.
            errors.raise_()


def generate_any_validator(**validators):
    return functools.partial(
        any_validator, validators=validators,
    )



import functools

from flex.datastructures import (
    ValidationList,
)
from flex.constants import OBJECT
from flex.validation.common import (
    generate_object_validator,
    apply_validator_to_object,
)
from .schema import (
    schema_validator,
)


schema_definitions_schema = {
    'type': OBJECT,
}

non_field_validators = ValidationList()
non_field_validators.add_validator(
    functools.partial(apply_validator_to_object, validator=schema_validator),
)

schema_definitions_validator = generate_object_validator(
    schema=schema_definitions_schema,
    non_field_validators=non_field_validators,
)
